#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <condefs.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#pragma hdrstop
//---------------------------------------------------------------------------
HINSTANCE hInstance;
HWND hed;


LPVOID lpMsgBuf;

char errbuf[1000];
char *err()
{ FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,0,GetLastError(),
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      errbuf,0,NULL);
  return errbuf;
}


// CRC code copied from the PNG specification.
unsigned long crc_table[256];
bool crc_table_computed = false;
void make_crc_table()
{ unsigned long c;
  int n, k;
  for (int n = 0; n < 256; n++)
  { unsigned long c = (unsigned long) n;
    for (int k = 0; k < 8; k++)
    { if (c&1) c = 0xedb88320L ^ (c>>1);
      else c = c>>1;
    }
    crc_table[n] = c;
  }
  crc_table_computed = true;
}
unsigned long update_crc(unsigned long crc, unsigned char *buf, int len)
{ unsigned long c = crc;
  int n;
  if (!crc_table_computed) make_crc_table();
  for (int n=0; n<len; n++)
  { c = crc_table[(c^buf[n]) & 0xff] ^ (c>>8);
  }
  return c;
}
unsigned long crc(unsigned char *buf, int len)
{ return update_crc(0xffffffffL, buf, len) ^ 0xffffffffL;
}




char *find_fn(char *pathfn)
{ char *fn1=strrchr(pathfn,'\\');
  char *fn2=strrchr(pathfn,':');
  if (fn1>fn2) return fn1+1;
  if (fn2>0) return fn2+1;
  return pathfn;
}

void change_ext(char *fn, char *ext)
{ char *dot = strrchr(fn,'.');
  if (dot!=0) *dot=0;
  strcat(fn,ext);
}


void FixPng(char *fn,char *msg)
{ strcpy(msg,fn);

  char *fnbase = find_fn(fn);
  double bbleft=0, bbbottom=0, bbright=0, bbtop=0;  // of EPS file
  unsigned int pixwidth=-1, pixheight=-1; // of PNG file

  // First, look for the EPS file, or BB, or PNG.BB
  HANDLE hfeps, hmeps; char *peps;
  char bbfn[MAX_PATH];
  strcpy(bbfn,fn); change_ext(bbfn,".eps");
  char *bbbase = find_fn(bbfn);
  hfeps = CreateFile(bbfn,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
  if (hfeps==INVALID_HANDLE_VALUE)
  { change_ext(bbfn,".bb");
    hfeps = CreateFile(bbfn,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
    if (hfeps==INVALID_HANDLE_VALUE)
    { change_ext(bbfn,".png.bb");
      hfeps = CreateFile(bbfn,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
      if (hfeps==INVALID_HANDLE_VALUE)
      { sprintf(msg,"! %s -- couldn't find .eps .bb .png.bb -- %s",fnbase,err());
        return;
      }
    }
  }
  hmeps = CreateFileMapping(hfeps,NULL,PAGE_READONLY,0,0,NULL);
  if (hmeps==NULL)
  { sprintf(msg,"! %s -- couldn't open BB filemapping -- %s",bbbase,err());
    CloseHandle(hfeps);
    return;
  }
  peps = (char*)MapViewOfFile(hmeps,FILE_MAP_READ,0,0,0);
  if (peps==NULL)
  { sprintf(msg,"! %s -- couldn't get BB filelock -- %s",bbbase,err());
    CloseHandle(hmeps); CloseHandle(hfeps);
    return;
  }
  char *c=peps; bool gotbb=false;
  while (*c=='%')
  { while (*c=='%') c++;
    if (strncmp(c,"HiResBoundingBox:",strlen("HiResBoundingBox:"))==0)
    { int i=sscanf(c+strlen("HiResBoundingBox:"),"%lf %lf %lf %lf",&bbleft,&bbbottom,&bbright,&bbtop);
      if (i==4) {gotbb=true; break;}
      else
      { sprintf(msg,"! %s -- corrupt HiResBoundingBox",bbbase);
        UnmapViewOfFile(peps); CloseHandle(hmeps); CloseHandle(hfeps);
        return;
      }
    }
    if (strncmp(c,"BoundingBox:",strlen("BoundingBox:"))==0)
    { int i=sscanf(c+strlen("BoundingBox:"),"%lf %lf %lf %lf",&bbleft,&bbbottom,&bbright,&bbtop);
      if (i==4) gotbb=true;
      else
      { sprintf(msg,"! %s -- corrupt BoundingBox",bbbase);
        UnmapViewOfFile(peps); CloseHandle(hmeps); CloseHandle(hfeps);
        return;
      }
    }
    if (strncmp(c,"EndComments",strlen("EndComments"))==0) break;
    while (*c!=0 && *c!='\r' && *c!='\n') c++;
    while (*c=='\r' || *c=='\n') c++;
  }
  UnmapViewOfFile(peps);
  CloseHandle(hmeps);
  CloseHandle(hfeps);
  if (!gotbb)
  { sprintf(msg,"! %s -- %%BoundingBox not found",bbbase);
    return;
  }
  if (bbright-bbleft==0 || bbtop-bbbottom==0)
  { sprintf(msg,"! %s -- empty bounding box",bbbase);
    return;
  }


  HANDLE hfpng, hffix, hmpng, hmfix; void *ppng, *pfix;
  bool usedexpansion=false;

  // Now, open the PNG file, and our own file for writing
  char outfn[MAX_PATH];
  strcpy(outfn,fn); change_ext(outfn,".png.fix");
  char *outbase = find_fn(outfn);
  hfpng = CreateFile(fn,GENERIC_READ,0,NULL,OPEN_EXISTING,0,NULL);
  if (hfpng==INVALID_HANDLE_VALUE)
  { sprintf(msg,"! %s -- couldn't open -- %s",fnbase,err());
    return;
  }
  DWORD hisize=0; DWORD size = GetFileSize(hfpng,&hisize);
  if (hisize!=0)
  { sprintf(msg,"! %s -- superlarge files not supported",fnbase);
    CloseHandle(hfpng);
    return;
  }
  hmpng = CreateFileMapping(hfpng,NULL,PAGE_READONLY,0,0,NULL);
  if (hmpng==NULL)
  { sprintf(msg,"! %s -- couldn't map file -- %s",fnbase,err());
    CloseHandle(hfpng);
    return;
  }
  ppng = (char*)MapViewOfFile(hmpng,FILE_MAP_READ,0,0,0);
  if (ppng==NULL)
  { sprintf(msg,"! %s -- couldn't lock file -- %s",fnbase,err());
    CloseHandle(hmpng); CloseHandle(hfpng);
    return;
  }
  hffix = CreateFile(outfn,GENERIC_READ|GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
  if (hffix==INVALID_HANDLE_VALUE)
  { sprintf(msg,"! %s -- couldn't open file for writing -- %s",outbase,err());
    UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng);
    return;
  }
  hmfix = CreateFileMapping(hffix,NULL,PAGE_READWRITE,0,size+12+9,NULL);
  if (hmfix==NULL)
  { sprintf(msg,"! %s -- couldn't map file for writing -- %s",outbase,err());
    CloseHandle(hffix); DeleteFile(outfn); UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng);
    return;
  }
  pfix =  MapViewOfFile(hmfix,FILE_MAP_ALL_ACCESS,0,0,0);
  if (pfix==NULL)
  { sprintf(msg,"! %s -- couldn't lock file for writing -- %s",outbase,err());
    CloseHandle(hmfix); CloseHandle(hffix); DeleteFile(outfn); UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng);
    return;
  }

  unsigned char *src = (unsigned char *)ppng;
  unsigned char *dst = (unsigned char *)pfix;
  if ( src[0]!=137 || src[1]!=80 || src[2]!=78 || src[3]!=71 || src[4]!=13
    || src[5]!=10 || src[6]!=26 || src[7]!=10)
  { sprintf(msg,"! %s -- isn't a PNG file -- invalid header",fnbase,err());
    UnmapViewOfFile(pfix); CloseHandle(hmfix); CloseHandle(hffix); DeleteFile(outfn); UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng); return;
  }
  CopyMemory(dst,src,8);
  src+=8; dst+=8;

  bool donephys=false;
  while (true)
  { bool mustcopy=true;
    unsigned long length = src[3]+256*(src[2]+256*(src[1]+256*src[0]));
    char type[5]; CopyMemory(type,src+4,4); type[4]=0;
    if (strcmp(type,"IHDR")==0)
    { pixwidth = src[11]+256*(src[10]+256*(src[9]+256*src[8]));
      pixheight = src[15]+256*(src[14]+256*(src[13]+256*src[12]));
    }
    if (strcmp(type,"pHYs")==0 || (strcmp(type,"IDAT")==0 && !donephys))
    { if (strcmp(type,"pHYs")==0 && length!=9)
      { sprintf(msg,"! %s -- current size information -- pHYs block not size 9",fnbase);
        UnmapViewOfFile(pfix); CloseHandle(hmfix); CloseHandle(hffix); DeleteFile(outfn); UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng); return;
      }
      unsigned long plen=9;
      dst[0]=(plen>>24)&255; dst[1]=(plen>>16)&255; dst[2]=(plen>>8)&255; dst[3]=(plen>>0)&255;
      dst[4]='p'; dst[5]='H'; dst[6]='Y'; dst[7]='s';
      if (pixwidth==-1 || pixwidth==0 || pixheight==-1 || pixheight==0)
      { sprintf(msg,"! %s -- empty image",fnbase);
        UnmapViewOfFile(pfix); CloseHandle(hmfix); CloseHandle(hffix); DeleteFile(outfn); UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng); return;
      }
      double widthpoints = (bbright-bbleft);
      double heightpoints = (bbtop-bbbottom);
      double widthmetres = widthpoints * 0.3527 / 1000.0;
      double heightmetres = heightpoints * 0.3527 / 1000.0;
      double dxppm = ((double)pixwidth)/widthmetres;
      double dyppm = ((double)pixheight)/heightmetres;
      dxppm = floor(dxppm+0.5);
      dyppm = floor(dyppm+0.5);
      unsigned long xppm = dxppm;
      unsigned long yppm = dyppm;
      dst[8]=(xppm>>24)&255; dst[9]=(xppm>>16)&255; dst[10]=(xppm>>8)&255; dst[11]=(xppm>>0)&255;
      dst[12]=(yppm>>24)&255; dst[13]=(yppm>>16)&255; dst[14]=(yppm>>8)&255; dst[15]=(yppm>>0)&255;
      dst[16]=1;
      unsigned long c = crc(dst+4, plen+4);
      dst[17]=(c>>24)&255; dst[18]=(c>>16)&255; dst[19]=(c>>8)&255; dst[20]=(c>>0)&255;
      char *verb;
      if (type=="pHYs")
      { verb="updated";
        dst += plen+12;
        src += plen+12;
        mustcopy=false;
      }
      else
      { verb="added";
        dst += plen+12;
        usedexpansion=true;
      }
      sprintf(msg,"%s -- %s size information -- %.0lfmm*%.0lfmm",fnbase,verb,widthmetres*1000,heightmetres*1000);
      donephys=true;
    }

    if (mustcopy)
    { unsigned long check = crc(src+4, length+4);
      unsigned long real = src[3+8+length]+256*(src[2+8+length]+256*(src[1+8+length]+256*src[0+8+length]));
      if (check!=real)
      { sprintf(msg,"! %s -- currupt png file -- invalid checksum in %s chunk",fnbase,type);
        UnmapViewOfFile(pfix); CloseHandle(hmfix); CloseHandle(hffix); DeleteFile(outfn); UnmapViewOfFile(ppng); CloseHandle(hmpng); CloseHandle(hfpng); return;
      }
      CopyMemory(dst,src,length+12);
      src+=length+12; dst+=length+12;
    }
    if (strcmp(type,"IEND")==0) break;
  }

  UnmapViewOfFile(ppng);
  CloseHandle(hmpng);
  CloseHandle(hfpng);
  UnmapViewOfFile(pfix);
  CloseHandle(hmfix);
  if (!usedexpansion)
  { DWORD size = GetFileSize(hffix,0);
    size = size-12-9;
    SetFilePointer(hffix,size,0,FILE_BEGIN);
    SetEndOfFile(hffix);
  }
  CloseHandle(hffix);
  DeleteFile(fn);
  RenameFile(outfn,fn);
}


LRESULT CALLBACK PlainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{ switch (msg)
  { case WM_CREATE:
    { hed = CreateWindow("EDIT","",WS_CHILD|WS_VISIBLE|WS_HSCROLL|WS_VSCROLL|ES_LEFT|ES_MULTILINE,
                    0,0,100,100,hwnd,NULL,hInstance,0);
      char *c = "PNG Bounding Box\r\n"
                "(for in case your graphics program\r\n"
                "doesn't support 'physical sizes')\r\n\r\n"
                "Drop your .PNG files into me.\r\n"
                "I will look for .EPS, .BB or .PNG.BB\r\n"
                "and set the physical size of the .PNG\r\n"
                "from them. The EPS/BB files must be\r\n"
                "text files including a line e.g.\r\n"
                "%%BoundingBox: 0 0 138 144\r\n"
                "or\r\n"
                "%%HiResBoundingBox: 0 0 138.2 144\r\n";
      SendMessage(hed,WM_SETTEXT,0,(LPARAM)c);
      DragAcceptFiles(hwnd,TRUE);
    } break;
    case WM_DROPFILES:
    { SendMessage(hed,WM_SETTEXT,0,(LPARAM)"");
      HDROP hdrop = (HDROP)wParam;
      int num = DragQueryFile(hdrop,0xFFFFFFFF,NULL,0);
      char msg[8000];
      for (int i=0; i<num; i++)
      { int len = DragQueryFile(hdrop,i,0,0);
        char *c=new char[len+1];
        DragQueryFile(hdrop,i,c,len+1);
        FixPng(c,msg);
        strcat(msg,"\r\n");
        SendMessage(hed,EM_REPLACESEL,FALSE,(LPARAM)msg);
        SendMessage(hed,EM_SETSEL,-1,-1);
      }
      DragFinish(hdrop);
    } break;
    case WM_PAINT:
    { PAINTSTRUCT ps; BeginPaint(hwnd,&ps);
      EndPaint(hwnd,&ps);
    } break;
    case WM_SIZE:
    { RECT rc; GetClientRect(hwnd,&rc);
      MoveWindow(hed,0,0,rc.right,rc.bottom,TRUE);
    } break;
    case WM_CLOSE:
    { DestroyWindow(hwnd);
    } break;
    case WM_DESTROY:
    { DragAcceptFiles(hwnd,FALSE);
      PostQuitMessage(0);
    } break;
  }
  return DefWindowProc(hwnd, msg, wParam, lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  //
  WNDCLASSEX wcex; ZeroMemory(&wcex,sizeof(wcex)); wcex.cbSize = sizeof(WNDCLASSEX);
  BOOL res=GetClassInfoEx(hInstance,"PlainClass",&wcex);
  if (!res)
  { wcex.style = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc = (WNDPROC)PlainWndProc;
    wcex.cbClsExtra = 0;
    wcex.cbWndExtra = 0;
    wcex.hInstance = hInstance;
    wcex.hIcon = NULL;
    wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcex.hbrBackground = NULL;
    wcex.lpszMenuName	= NULL;
    wcex.lpszClassName = "PlainClass";
    wcex.hIconSm = NULL;
    ATOM res=RegisterClassEx(&wcex);
    if (res==0) {MessageBox(NULL,"Failed to register class","Error",MB_OK); return 0;}
  }
  //
  HWND hwnd = CreateWindowEx(0,"PlainClass", "PNG Bounding Box", WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, 400, 400, NULL, NULL, hInstance, NULL);
  if (hwnd==NULL) {MessageBox(NULL,"Failed to create window","Error",MB_OK); return 0;}
  ShowWindow(hwnd,SW_SHOW);
  //
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0))
  { TranslateMessage(&msg);
	DispatchMessage(&msg);
  }
  return msg.wParam;
}

 
